from typing import TypedDict
from pathlib import Path
import functools
import os

from arc.color import fg, effects
from .types import DirectorySchema
from .config import config
from .icons import default, filetype_icon, Icon
from . import util

GLOB = "/*"
EXCLUDED_DIRS = (".git", "node_modules")
browse = config.browse


class BrowseItem(TypedDict):
    path: str
    icon: Icon


class PathError(Exception):
    ...


class BrowseList:
    def __init__(self, entries: DirectorySchema):
        self.entries = entries
        self.items: dict[str, BrowseItem] = {}

    def discover(self):
        for entry in self.entries:
            if isinstance(entry, str):
                # pylint: disable=no-value-for-parameter
                self.handle_single_path(entry)
            elif isinstance(entry, dict):
                self.handle_hierarchy(entry["base"], entry["subs"])

    def handle_single_path(self, directory: str, prefix: str = None):
        directory = util.resolve(directory)
        prefix = prefix or directory
        path = Path(directory)

        if path.as_posix().endswith(GLOB):
            self.handle_glob(path)

        elif not path.exists():
            raise PathError(f"{path} is not a valid file or directory")

        elif path.is_file():
            self.add_entry(path, prefix)

        elif path.is_dir():
            for sub in path.iterdir():
                self.add_entry(sub, prefix)

    def handle_hierarchy(self, base: str, subs: list[str]):
        paths = [f"{base}/{sub.strip('/')}" for sub in subs]
        # Quick hack to fix globbed paths not getting
        # matched in the below if
        removed_globs = [path.rstrip(GLOB) for path in paths]

        for path in os.listdir(base):
            abs_path = Path(base) / path
            if (
                abs_path.as_posix() not in removed_globs
            ) or browse.include_parent_folder:
                self.add_entry(abs_path, base)

        for path in paths:
            self.handle_single_path(path, base)

    def handle_glob(self, globbed_path: Path):
        path = Path(globbed_path.as_posix().removesuffix(GLOB))
        path_str = path.as_posix()
        subs = [
            entry.as_posix().removeprefix(path_str).removeprefix("/")
            for entry in path.iterdir()
            if (browse.show_hidden or not entry.name.startswith("."))
            and entry.name not in EXCLUDED_DIRS
        ]

        self.handle_hierarchy(path_str, subs)

    def add_entry(self, path: Path, prefix: str):
        if (path.name.startswith(".") and not browse.show_hidden) or (
            path.is_file() and not browse.show_files
        ):
            return

        name = self.get_unique_path(path, prefix)
        if path.is_file():
            icon = self.get_icon(path.suffix)
        else:
            icon = self.get_icon("directory")

        self.items[name] = {
            "path": path.as_posix(),
            "icon": icon,
        }

    def get_unique_path(self, path: Path, prefix: str):
        string = path.as_posix().removeprefix(prefix).lstrip("/")
        prefix_path = Path(prefix)

        # Generates the shortest path that is unique
        # given defined entry points
        parts_len = len(prefix_path.parts)
        selection_start = parts_len - 1
        while string in self.items:
            string = (
                "/".join(prefix_path.parts[selection_start:parts_len]) + "/" + string
            )

            if selection_start <= 0:
                raise PathError(
                    f"Was unable to create a unique name for {fg.YELLOW}{path}{effects.CLEAR} "
                    "this is likely because multiple entries contain the given path"
                )

            if string in self.items:
                selection_start -= 1

        return string

    @staticmethod
    @functools.lru_cache()
    def get_icon(filetype: str) -> Icon:
        """Returns the icon associated with a specific file type"""
        return filetype_icon(filetype.lstrip(".")) or default



# pylint: disable=inherit-non-class
from __future__ import annotations
from typing import List, Union, TypedDict, Optional
from enum import Enum


class OpenContext(Enum):
    BROWSE = "Browse"
    QUICK_LAUNCH = "QuickLaunch"


class BaseSub(TypedDict):
    base: str
    subs: list[str]


DirectorySchema = List[Union[str, BaseSub]]


class QuickLaunchEntry(TypedDict):
    path: str
    icon: str
    tags: Optional[list[str]]



from __future__ import annotations
from typing import Mapping, Any, TypeVar, Generic
from abc import ABC, abstractmethod

from ..config import Config


T = TypeVar("T")
V = TypeVar("V")


class UIProvider(Generic[T, V], ABC):
    def __init__(self, items: T, config: Config):
        self.items = items
        self.config = config
        self._display_items = self.format_items(items)

    @abstractmethod
    def run(self) -> V | None:
        """Run the UI Interface. Return the selected value"""

    def format_items(self, items: T) -> Mapping[str, str]:
        """Format the input items into a dictionary
        where the key will be displayed in the UI
        and the value should be returned to the caller
        """
        return {key: key for key in items}  # type: ignore

    def parse_output(self, output: Any) -> str:
        """Called post-execution, used to do any
        nessecary parsing of the chosen answer

        If no parsing is nessecary, there is no need
        to implement this method
        """
        return output



import os
import functools
from typing import Any
from contextlib import contextmanager

from arc.utils import logger
import ujson


def resolve(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path))


def pango_span(content, **kwargs):
    return (
        "<span "
        + " ".join(f'{key}="{value}"' for key, value in kwargs.items())
        + f">{content}</span>"
    )


def catch(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise Cache.CacheError(f"Cache {func.__name__} failed") from e

    return wrapper


class Cache:
    cache_path = resolve("~/.cache/venture-cache")

    class CacheError(Exception):
        ...

    def __init__(self, config):
        self.config = config
        self.meta: dict[str, Any] = {}
        self.data: dict[str, str] = {}

    @catch
    def read(self):
        logger.debug("Reading %s", self.cache_path)
        data: dict = self.attempt_load()
        self.meta = data.get("meta", {})
        self.data = data.get("data", {})
        return self.data

    @catch
    def write(self, data: dict):
        logger.debug("Writing %s", self.cache_path)
        self.data |= data
        self.meta["checksum"] = self.config.checksum
        to_dump = {"data": self.data, "meta": self.meta}
        with open(self.cache_path, "w") as f:
            f.write(ujson.dumps(to_dump, ensure_ascii=False))

    def attempt_load(self):
        try:
            with open(self.cache_path, "r") as f:
                return ujson.loads(f.read())
        except (ValueError, FileNotFoundError):
            return {}

    def valid(self):
        return self.exists() and self.config.checksum == self.meta.get("checksum")

    @classmethod
    def exists(cls):
        return os.path.isfile(cls.cache_path)


def confirm(message: str):
    result = {
        "y": True,
        "n": False,
    }
    print(f"{message} [Y/N]")
    while True:
        user_input = input("> ").lower()
        if user_input in result:
            return result[user_input]
        print("Not understood, please type 'y' or 'n'")


@contextmanager
def safe_write(file_name: str):
    """Save the contents of the file into memory, then
    yield the file object. If an error occurs, writes the contents
    back into the file.
    """
    with open(file_name, "r") as f:
        contents = f.read()

    file = open(file_name, "w")

    try:
        yield file
    except:
        file.write(contents)
        raise
    finally:
        file.close()
